' **********************************************************************
' Solution to a system of nonlinear equations by Broyden's method.
'
' Example (from Numerical Recipes Example Book):
'
' F(X, Y) = X^2 + Y^2 - 2 = 0
' G(X, Y) = EXP(X - 1) + Y^3 - 2 = 0
'
'               ( 2 * X          2 * Y   )
' Jacobian: D = (                        )
'               ( EXP(X - 1)     3 * Y^2 )
'
' True Solution is at (1, 1)
' **********************************************************************

#INCLUDE "fbmath.bi"

DECLARE SUB Equations(X() AS DOUBLE, F() AS DOUBLE)

' **********************************************************************
' Define number of variables, number of iterations, and precision
' **********************************************************************

CONST Nvar    = 2     ' Number of variables
CONST MaxIter = 1000  ' Max number of iterations
CONST Tol     = 1D-6  ' Required precision (must be > Sqrt(MachEp))

' **********************************************************************
' Define starting point
' **********************************************************************

' Variables: X(1) = X, X(2) = Y
DIM AS DOUBLE X(1 TO Nvar) = {2, 0.5}

' **********************************************************************
' Define the system of equations to be solved
' **********************************************************************

SUB Equations (X() AS DOUBLE, F() AS DOUBLE)

  DIM AS DOUBLE X1p2, X2p2, X2p3

  X1p2 = X(1) * X(1)
  X2p2 = X(2) * X(2)
  X2p3 = X(2) * X2p2

  F(1) = X1p2 + X2p2 - 2
  F(2) = EXP(X(1) - 1) + X2p3 - 2
END SUB

' **********************************************************************
' Main program
' **********************************************************************

DIM AS DOUBLE  F(1 TO Nvar)  ' Functions
DIM AS INTEGER I, J          ' Loop variables

Broyden @Equations, X(), MaxIter, Tol, F()

IF MathErr() = OptNonConv THEN
  PRINT "Non-convergence!"
  STOP
END IF

CLS

PRINT "Solution to nonlinear equation system (Broyden's method)"
PRINT "--------------------------------------------------------"
PRINT

PRINT "Solution vector:" : PRINT

FOR I = 1 TO Nvar
  PRINT USING "X(#) = ####.######"; I; X(I)
NEXT I

PRINT : PRINT "Function values:" : PRINT

FOR I = 1 TO Nvar
  PRINT USING "F(#) = ####.######"; I; F(I)
NEXT I

SLEEP
