' **********************************************************************
' Solution to a system of nonlinear equations by the Newton-Raphson
' method.
'
' Example (from Numerical Recipes Example Book):
'
' F(X, Y) = X^2 + Y^2 - 2 = 0
' G(X, Y) = EXP(X - 1) + Y^3 - 2 = 0
'
'               ( 2 * X          2 * Y   )
' Jacobian: D = (                        )
'               ( EXP(X - 1)     3 * Y^2 )
'
' True Solution is at (1, 1)
' **********************************************************************

#INCLUDE "fbmath.bi"

' **********************************************************************
' Define number of variables, number of iterations, and precision
' **********************************************************************

CONST Nvar    = 2     ' Number of variables
CONST MaxIter = 1000  ' Max number of iterations
CONST Tol     = 1D-6  ' Required precision (must be > Sqrt(MachEp))

' **********************************************************************
' Define starting point
' **********************************************************************

' Variables: X(1) = X, X(2) = Y
DIM AS DOUBLE X(1 TO Nvar) = {2, 0.5}

' **********************************************************************
' Define the system of equations to be solved
' **********************************************************************

SUB Equations (X() AS DOUBLE, F() AS DOUBLE)

  DIM AS DOUBLE X1p2, X2p2, X2p3

  X1p2 = X(1) * X(1)
  X2p2 = X(2) * X(2)
  X2p3 = X(2) * X2p2

  F(1) = X1p2 + X2p2 - 2
  F(2) = EXP(X(1) - 1) + X2p3 - 2
END SUB

' **********************************************************************
' Define the subroutine which computes the jacobian of the system.
' It is recommended to use analytical derivatives whenever possible.
' Otherwise you can use the alternative code provided below.
' **********************************************************************

SUB Jacobian (X() AS DOUBLE, D() AS DOUBLE)

  D(1, 1) = 2 * X(1)
  D(1, 2) = 2 * X(2)
  D(2, 1) = EXP(X(1) - 1)
  D(2, 2) = 3 * X(2) * X(2)
END SUB

' **********************************************************************
' Alternative code to be used if the analytical derivatives are not
' available.
' **********************************************************************

'SUB Jacobian (X() AS DOUBLE, D() AS DOUBLE)
'
' Define the relative increment used to compute numerical derivatives
' (should not be lower than MachEp^1/2 which is about 1.5D-8
'
'  CONST Eta = 1D-6
'
'  NumJacobian @Equations, X(), Eta, D()
'END SUB

' **********************************************************************
' Main program
' **********************************************************************

DIM AS DOUBLE  F(1 TO Nvar)             ' Functions
DIM AS DOUBLE  D(1 TO Nvar, 1 TO Nvar)  ' Jacobian
DIM AS INTEGER I, J                     ' Loop variables

NewtEqs @Equations, @Jacobian, X(), MaxIter, Tol, F()

SELECT CASE MathErr()
  CASE OptNonConv
    PRINT "Non-convergence!"
    STOP
  CASE OptSing
    PRINT "Singular Jacobian matrix!"
    STOP
END SELECT

CLS

PRINT "Solution to nonlinear equation system (Newton-Raphson method)"
PRINT "-------------------------------------------------------------"
PRINT

PRINT "Solution vector:": PRINT

FOR I = 1 TO Nvar
  PRINT USING "X(#) = ####.######"; I; X(I)
NEXT I

PRINT : PRINT "Function values:" : PRINT

FOR I = 1 TO Nvar
  PRINT USING "F(#) = ####.######"; I; F(I)
NEXT I

SLEEP

