' ******************************************************************
' Optimization by Simulated Annealing
' ******************************************************************

#INCLUDE "fbmath.bi"

' ******************************************************************
' Subprograms
' ******************************************************************

FUNCTION Square(X AS DOUBLE) AS DOUBLE
  Square = X * X
END FUNCTION

FUNCTION Func1(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Example taken from 'Numerical Recipes'
' True minimum is at (-2.0, +/-0.89442719)
' ------------------------------------------------------------------

  DIM AS DOUBLE A, AA, B, BB

  A = X(2) * X(2) * (3 - X(1)) - X(1) * X(1) * (3 + X(1))
  B = 2 + X(1)
  AA = A * A
  BB = B * B
  Func1 = 10 * AA + BB / (1 + BB)
END FUNCTION

FUNCTION Func2(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Example taken from 'Numerical Recipes'
' True minimum is at (0, 0, 0, 0), F = 1.0
' ------------------------------------------------------------------

  CONST N_VAR = 4, RAD = 0.3, AUG  = 2

  STATIC AS DOUBLE Wid(1 TO N_VAR) = {1, 3, 10, 30}

  DIM AS INTEGER J
  DIM AS DOUBLE  Q, R, Rad2, Sumd, Sumr

  Sumd = 0
  Sumr = 0
  Rad2 = Square(RAD)

  FOR J = 1 TO N_VAR
    Q = X(J) * Wid(J)
    IF Q >= 0 THEN R = INT(Q + 0.5) ELSE R = INT(Q - 0.5)
    Sumr = Sumr + Square(Q)
    Sumd = Sumd + Square(Q - R)
  NEXT J

  IF Sumd > Rad2 THEN
    Func2 = 1 + Sumr * (1 + AUG)
  ELSE
    Func2 = 1 + Sumr * (1 + AUG * Sumd / Rad2)
  END IF
END FUNCTION

FUNCTION Func3(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Rosenbrock function.
'
' True minimum is at (1, 1), F = 0
'
' Ref: H. Rosenbrock, Comput. J., 1960, 3, 175
' ------------------------------------------------------------------

  Func3 = 100 * Square(X(2) - Square(X(1))) + Square(1 - X(1))
END FUNCTION

FUNCTION Func4(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Powell function.
'
' True minimum is at (0, 0, 0, 0), F = 0
'
' Ref: M.J.D. Powell, Comput. J., 1962, 5, 147
' ------------------------------------------------------------------

  Func4 = Square(X(1) + 10 * X(2)) + 5 * Square(X(3) - X(4)) + _
          Square(Square(X(2) - 2 * X(3))) + 10 * Square(Square(X(1) - X(4)))
END FUNCTION

FUNCTION Func5(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Another Powell function.
'
' Multiple minima at x1 = x2 = x3 = +/- Sqrt(4*n+1), n integer, F = -3
'
' Ref: M.J.D. Powell, Comput. J., 1964, 7, 155
'
' NB: The original reference maximizes F. Here we shall minimize -F.
' ------------------------------------------------------------------

  Func5 = -1 / (1 + Square(X(1) - X(2))) - SIN(PiDiv2 * X(2) * X(3)) _
          - Expo(- Square((X(1) + X(3)) / X(2) - 2))
END FUNCTION

FUNCTION Func6(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Fletcher & Powell function.
'
' True minimum is at (1, 0, 0), F = 0
'
' Ref: R. Fletcher & M.J.D. Powell, Comput. J., 1964, 7, 155
' ------------------------------------------------------------------

  DIM AS DOUBLE R, Theta

  R = Pythag(X(1), X(2))
  Theta = ATAN2(X(2), X(1)) / TwoPi
  Func6 = 100 * (Square(X(3) - 10 * Theta) + Square(R - 1)) + Square(X(3))
END FUNCTION

FUNCTION Func7(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Colville function (Extension of Rosenbrock function)
'
' True minimum is at (1, 1, 1, 1), F = 0
'
' Ref: R. J. Van Iwaarden, PhD Thesis, U. Denver, 1996
' ------------------------------------------------------------------

  Func7 = 100 * Square(X(2) - Square(X(1))) + Square(1 - X(1)) + _
           90 * Square(X(4) - Square(X(3))) + Square(1 - X(3)) + _
           10.1 * ((Square(X(2) - 1) + Square(X(4) - 1))) +      _
           19.8 * (X(2) - 1) * (X(4) - 1)
END FUNCTION

FUNCTION Func8(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Griewank function.
'
' True minimum is at (0, 0), F = 0
'
' Ref: R. J. Van Iwaarden, PhD Thesis, U. Denver, 1996
' ------------------------------------------------------------------

  Func8 = (Square(X(1)) + Square(X(2))) / 200 _
          - COS(X(1)) * COS(X(2) / Sqrt2) + 1
END FUNCTION

FUNCTION Func9(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Chichinadze function.
'
' True minimum is at (5.90133, 0.5), F = -43.3159
'
' Ref: R. J. Van Iwaarden, PhD Thesis, U. Denver, 1996
' ------------------------------------------------------------------

  CONST FivePi   = 15.707963267948966193   ' 5 * Pi
  CONST InvSqrt5 = 0.44721359549995793928  ' 1 / Sqrt(5)

  Func9 = X(1) * (X(1) - 12) + 11 _
             + 10 * COS(PiDiv2 * X(1)) + 8 * SIN(FivePi * X(1)) _
             - InvSqrt5 * Expo(- 0.5 * Square(X(2) - 0.5))
END FUNCTION

FUNCTION Func10(X() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Rastrigin function.
'
' True minimum is at (0, 0), F = -2
'
' Ref: R. J. Van Iwaarden, PhD Thesis, U. Denver, 1996
' ------------------------------------------------------------------

  Func10 = Square(X(1)) + Square(X(2)) - COS(12 * X(1)) - COS(18 * X(2))
END FUNCTION

' ******************************************************************
' Main program
' ******************************************************************

CONST NFunc   = 10  ' Number of functions
CONST MaxNVar =  4  ' Maximum number of variables

' Function names
DIM FuncName(1 TO NFunc) AS STRING*70 => _
{"Numerical Recipes Example 1: Minimum at (-2.0, +/-0.89442719), F = 0 ", _
 "Numerical Recipes Example 2: Minimum at (0, 0, 0, 0), F = 1          ", _
 "Rosenbrock function: Minimum at (1, 1), F = 0                        ", _
 "Powell function: Minimum at (0, 0, 0, 0), F = 0                      ", _
 "Another Powell function: Minimum at x1=x2=x3= +/- Sqrt(4*n+1), F = -3", _
 "Fletcher & Powell function: Minimum at (1, 0, 0), F = 0              ", _
 "Colville function: Minimum at (1, 1, 1, 1), F = 0                    ", _
 "Griewank function: Minimum at (0, 0), F = 0                          ", _
 "Chichinadze function: Minimum at (5.90133, 0.5), F = -43.3159        ", _
 "Rastrigin function: Minimum at (0, 0), F = -2                        "}

' Number of variables
DIM Nvar(1 TO NFunc) AS INTEGER = {2, 4, 2, 4, 3, 3, 4, 2, 2, 2}

' Array of functions
DIM Func(1 TO NFunc) AS FUNCTION(X() AS DOUBLE) AS DOUBLE = _
{@Func1, @Func2, @Func3, @Func4, @Func5, _
 @Func6, @Func7, @Func8, @Func9, @Func10}

DIM AS DOUBLE  Fmin  ' Function value at minimum
DIM AS INTEGER I, J  ' Loop variable

FOR I = 1 TO NFunc
  ' Dimension arrays
  REDIM AS DOUBLE X(1 TO Nvar(I))
  REDIM AS DOUBLE Xmin(1 TO Nvar(I))
  REDIM AS DOUBLE Xmax(1 TO Nvar(I))

  ' Pick starting point
  FOR J = 1 TO Nvar(I)
    Xmin(J) = -10
    Xmax(J) =  10
    X(J) = Xmin(J) + RanGen3 * (Xmax(J) - Xmin(J))
  NEXT J

  ' Approximate global minimum with simulated annealing
  SimAnn(Func(I), X(), Xmin(), Xmax(), Fmin)

  ' Display results
  PRINT FuncName(I)
  PRINT
  FOR J = 1 TO Nvar(I)
    PRINT USING "X(#) = ##.####^^^^"; J; X(J)
  NEXT J
  PRINT
  PRINT USING "F    = ##.####^^^^"; Fmin
  PRINT

  SLEEP
NEXT I

