' ******************************************************************
' Weighted linear regression
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE MatOk 0
' No error

#DEFINE MatSing -2
' Quasi-singular matrix

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the last function evaluation

' ******************************************************************

SUB WLinFit(X() AS DOUBLE, _
            Y() AS DOUBLE, _
            S() AS DOUBLE, _
            B() AS DOUBLE, _
            V() AS DOUBLE)
' ------------------------------------------------------------------
' Weighted linear regression: Y = B(0) + B(1) * X
' ------------------------------------------------------------------
' Input parameters:  X(), Y() = point coordinates
'                    S()      = standard deviations
' Output parameters: B()      = model parameters
'                    V()      = inverse matrix
' ------------------------------------------------------------------
' The function MathErr() returns one of the following codes:
'   MatOk   = no error
'   MatSing = singular matrix
' ------------------------------------------------------------------

  DIM AS DOUBLE  W, WX, SW, SWX, SWY, SWX2, SWXY, D
  DIM AS INTEGER K, Lb, Ub

  Lb = LBOUND(X)
  Ub = UBOUND(X)

  ErrCode = MatOk

  SW   = 0
  SWX  = 0
  SWY  = 0
  SWX2 = 0
  SWXY = 0

  FOR K = Lb TO Ub
    IF S(K) <= 0 THEN ErrCode = MatSing : EXIT SUB
    W = 1 / (S(K) * S(K))
    WX = W * X(K)
    SW = SW + W
    SWX = SWX + WX
    SWY = SWY + W * Y(K)
    SWX2 = SWX2 + WX * X(K)
    SWXY = SWXY + WX * Y(K)
  NEXT K

  D = SW * SWX2 - SWX * SWX

  IF D <= 0 THEN ErrCode = MatSing : EXIT SUB

  Lb = LBOUND(B)
  Ub = UBOUND(B)
  
  V(Lb, Lb) = SWX2 / D
  V(Lb, Ub) = - SWX / D
  V(Ub, Lb) = V(Lb, Ub)
  V(Ub, Ub) = SW / D

  B(Lb) = V(Lb, Lb) * SWY + V(Lb, Ub) * SWXY
  B(Ub) = V(Ub, Lb) * SWY + V(Ub, Ub) * SWXY
END SUB
