' ******************************************************************
' Weighted multiple linear regression
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE MatSing -2
' Quasi-singular matrix

#DEFINE MatErrDim -3
' Non-compatible dimensions 

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the last function evaluation

' ------------------------------------------------------------------
' External function
' ------------------------------------------------------------------

DECLARE SUB LinEq(      A() AS DOUBLE, _
                        B() AS DOUBLE, _
                  BYREF Det AS DOUBLE)
' Solves the linear system A * X = B

' ******************************************************************

SUB WMulFit(X() AS DOUBLE, _
            Y() AS DOUBLE, _
            S() AS DOUBLE, _
            B() AS DOUBLE, _
            V() AS DOUBLE)
' ------------------------------------------------------------------
' Weighted multilinear regression
' ------------------------------------------------------------------
' Input : X() = Matrix of independent variables
'         Y() = Vector of dependent variable
'         S() = Standard deviations
' Output: B() = Vector of regression parameters
'         V() = Inverse matrix of normal equations
' ------------------------------------------------------------------

  DIM AS INTEGER Lb, Ub, LbY, UbY, I, J, K
  DIM AS DOUBLE  W, WX, Det

  Lb = LBOUND(B)
  Ub = UBOUND(B)

  LbY = LBOUND(Y)
  UbY = UBOUND(Y)

  IF LBOUND(X, 1) <> LbY OR UBOUND(X, 1) <> UbY OR _
     LBOUND(X, 2) <> Lb  OR UBOUND(X, 2) <> Ub  OR _
     LBOUND(V, 1) <> Lb  OR UBOUND(V, 1) <> Ub  OR _
     LBOUND(V, 2) <> Lb  OR UBOUND(V, 2) <> Ub     _
  THEN
    ErrCode = MatErrDim
    EXIT SUB
  END IF

  ' Initialize
  FOR I = Lb TO Ub
    FOR J = I TO Ub
      V(I, J) = 0
    NEXT J
    B(I) = 0
  NEXT I
  
  ' Compute system of normal equations
  FOR K = LbY TO UbY
    IF S(K) <= 0 THEN ErrCode = MatSing : EXIT SUB
    W = 1 / (S(K) * S(K))
    FOR I = Lb TO Ub
      WX = W * X(K, I)
      FOR J = I TO Ub
        V(I, J) = V(I, J) + WX * X(K, J)  ' X'WX
      NEXT J
      B(I) = B(I) + WX * Y(K)             ' X'WY
    NEXT I
  NEXT K

  ' Fill in symmetric matrix
  FOR I = Lb + 1 TO Ub
    FOR J = Lb TO I - 1
      V(I,J) = V(J,I)
    NEXT J
  NEXT I
        
  LinEq V(), B(), Det
END SUB
