' ******************************************************************
' Weighted polynomial regression
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE MatSing -2
' Singular matrix

#DEFINE MatErrDim -3
' Non-compatible dimensions

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the last function evaluation

' ------------------------------------------------------------------
' External function
' ------------------------------------------------------------------

DECLARE SUB LinEq(      A() AS DOUBLE, _
                        B() AS DOUBLE, _
                  BYREF Det AS DOUBLE)
' Solves the linear system A * X = B

' ******************************************************************

SUB WPolFit(X() AS DOUBLE,  _
            Y() AS DOUBLE,  _
            S() AS DOUBLE,  _
            B() AS DOUBLE,  _
            V() AS DOUBLE)
' ------------------------------------------------------------------
' Weighted polynomial regression: Y = B(0) + B(1) * X + B(2) * X^2 + ...
' ------------------------------------------------------------------
' Input parameters:  X(), Y() = point coordinates
'                    S()      = standard deviations
' Output parameters: B()      = model parameters
'                    V()      = inverse matrix
' ------------------------------------------------------------------
' The function MathErr() returns one of the following codes:
'   MatOk     = no error
'   MatSing   = singular matrix
'   MatErrDim = non-compatible dimensions
' ------------------------------------------------------------------

  DIM AS INTEGER I, I1, J, K, Lb, Lb1, LbX, Ub, Ub1, UbX
  DIM AS DOUBLE  W, WXI, Det

  Lb = LBOUND(B)
  Ub = UBOUND(B)

  LbX = LBOUND(X)
  UbX = UBOUND(X)

  IF LBOUND(Y)    <> LbX OR UBOUND(Y)    <> UbX OR _
     LBOUND(V, 1) <> Lb  OR UBOUND(V, 1) <> Ub  OR _
     LBOUND(V, 2) <> Lb  OR UBOUND(V, 2) <> Ub     _
  THEN
    ErrCode = MatErrDim
    EXIT SUB
  END IF

  ' Initialize
  FOR I = Lb TO Ub
    FOR J = Lb TO Ub
      V(I, J) = 0
    NEXT J
    B(I) = 0
  NEXT I
  
  Lb1 = Lb + 1    
  FOR K = LbX TO UbX
    IF S(K) <= 0 THEN ErrCode = MatSing : EXIT SUB 
    W = 1 / (S(K) * S(K))
    WXI = W * X(K)                 ' w.x^i
    V(Lb, Lb) = V(Lb, Lb) + W
    B(Lb) = B(Lb) + W * Y(K)
    V(Lb, Lb1) = V(Lb, Lb1) + WXI
    B(Lb1) = B(Lb1) + WXI * Y(K)
    FOR I = Lb + 2 TO Ub           
      WXI = WXI * X(K)                 
      V(Lb, I) = V(Lb, I) + WXI    ' First line of matrix: w --> w.x^d
      B(I) = B(I) + WXI * Y(K)     ' Constant vector: w.y --> w.x^d.y
    NEXT I                           
    FOR I = Lb1 TO Ub              
      WXI = WXI * X(K)              
      V(I, Ub) = V(I, Ub) + WXI    ' Last column of matrix: w.x^d --> w.x^2d
    NEXT I
  NEXT K

 ' Fill lower matrix
  Ub1 = Ub - 1
  FOR I = Lb1 TO Ub    
    I1 = I - 1  
    FOR J = Lb TO Ub1
      V(I, J) = V(I1, J + 1)
    NEXT J
  NEXT I

  ' Solve system
  LinEq V(), B(), Det
END SUB

