' ******************************************************************
' Newton-Raphson solver for nonlinear equation
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE OptOk 0  
' No error

#DEFINE OptNonConv -1  
' Non-convergence

#DEFINE OptSing -2  
' Singularity (null derivative)

' ------------------------------------------------------------------
' Machine-dependent constant
' ------------------------------------------------------------------

#DEFINE MachEp 2.220446049250313D-16  
' Floating point precision: 2^(-52)

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the last function evaluation

' ******************************************************************

SUB NewtEq(      Func    AS FUNCTION(X AS DOUBLE) AS DOUBLE,_
                 Deriv   AS FUNCTION(X AS DOUBLE) AS DOUBLE,_
           BYREF X       AS DOUBLE,_
                 MaxIter AS INTEGER,_ 
                 Tol     AS DOUBLE,_ 
           BYREF F       AS DOUBLE)
' ----------------------------------------------------------------------
' Solves a nonlinear equation by Newton's method
' ----------------------------------------------------------------------
' Input parameters  : Func    = function to be solved
'                     Deriv   = derivative
'                     X       = initial root
'                     MaxIter = maximum number of iterations
'                     Tol     = required precision
' ----------------------------------------------------------------------
' Output parameters : X = refined root
'                     F = function value
' ----------------------------------------------------------------------
' Possible results : OptOk      = no error
'                    OptNonConv = non-convergence
'                    OptSing    = singularity (null derivative)
' ----------------------------------------------------------------------

  DIM AS INTEGER Iter  ' Iteration count
  DIM AS DOUBLE  OldX  ' Old root
  DIM AS DOUBLE  D     ' Derivative
  DIM AS DOUBLE  Xtol  ' Tolerance
    
  ' Initialize function value
  F = Func(X)

  IF MaxIter < 1 THEN EXIT SUB
   
  Iter = 0
  
  DO
   ' Compute derivative
    D = Deriv(X)

    IF D = 0 THEN ErrCode = OptSing : EXIT SUB
          
    ' Prepare next iteration
    Iter = Iter + 1
    IF Iter > MaxIter THEN ErrCode = OptNonConv : EXIT SUB

    ' Save current root and compute new one
    OldX = X
    X = X - F / D
    F = Func(X)
    
    Xtol = Tol * ABS(X)
    IF Xtol < MachEp THEN Xtol = MachEp
  LOOP UNTIL ABS(OldX - X) < Xtol
  
  ErrCode = OptOk
END SUB

