' ******************************************************************
' DiGamma and TriGamma functions.
' Based on a contribution by Philip Fletcher (FLETCHP@WESTAT.com)
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE FOk 0
' No error

#DEFINE FSing -2
' Singularity

' ------------------------------------------------------------------
' Machine-dependent constant
' ------------------------------------------------------------------

#DEFINE MaxNum 1.797693134862315D+308
' Max. floating point number: 2^1024

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the latest function evaluation

' ******************************************************************

FUNCTION DiGamma (X AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Digamma calculates the Digamma or Psi function
' = d ( Ln ( Gamma ( X ) ) ) / dX
'
'  Reference:
'
'    J Bernardo,
'    Psi ( Digamma ) Function,
'    Algorithm AS 103,
'    Applied Statistics,
'    Volume 25, Number 3, pages 315-317, 1976.
'
'  Modified:
'
'    03 January 2000
'
'  Parameters:
'
'    Input, real X, the argument of the Digamma function.
'    0 < X.
'
'    Output, real Digamma, the value of the Digamma function at X.
' ------------------------------------------------------------------

  CONST C  = 20#
  CONST S  = 0.00001#
  CONST D1 = -0.57721566490153286061#  ' DiGamma(1)

  ' Sterling coefficient S(n) = B(n) / 2n
  ' where B(n) = Bernoulli number

  CONST S2  =  0.08333333333333333333#    ' B(2)/2
  CONST S4  = -0.83333333333333333333D-2  ' B(4)/4
  CONST S6  =  0.39682539682539682541D-2  ' B(6)/6
  CONST S8  = -0.41666666666666666666D-2  ' B(8)/8
  CONST S10 =  0.75757575757575757576D-2  ' B(10)/10
  CONST S12 = -0.21092796092796092796D-1  ' B(12)/12
  CONST S14 =  0.83333333333333333335D-1  ' B(14)/14
  CONST S16 = -0.44325980392156862745     ' B(16)/16

  ' Function argument must be positive.
 
  IF X <= 0 THEN
    ErrCode = FSing
    DiGamma = MaxNum
    EXIT FUNCTION
  END IF

  DIM AS DOUBLE Dg, P, R, Y

  ErrCode = FOk

  IF X = 1 THEN
    DiGamma = D1
    EXIT FUNCTION
  END IF

  ' Use approximation if argument <= S.

  IF X <= S THEN

    Dg = D1 - 1 / X

  ' Reduce the argument to Dg(X + N) where (X + N) >= C.

  ELSE
    Dg = 0
    Y = X

    DO WHILE Y < C
      Dg = Dg - 1 / Y
      Y = Y + 1
    LOOP

    ' Use Stirling's (actually de Moivre's) expansion if argument > C.

    R = 1 / (Y * Y)
    
    P = (((((((S16 * R + S14) * R + S12) * R + S10) * R + S8) * R + _
               S6) * R + S4) * R + S2) * R

    Dg = Dg + LOG(Y) - 0.5 / Y - P
  END IF

  DiGamma = Dg
END FUNCTION

FUNCTION TriGamma (X AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Trigamma calculates the Trigamma or Psi Prime function =
'  d**2 ( LOG ( GAMMA ( X ) ) ) / dX**2
'
'  Reference:
'
'    Algorithm As121 Appl. Statist. (1978) vol 27, no. 1
' ------------------------------------------------------------------

  CONST A    = 0.0001#
  CONST B    = 20
  CONST Zero = 0
  CONST One  = 1
  CONST Half = 0.5

  ' Bernoulli numbers

  CONST B2  =  0.1666666666666667#
  CONST B4  = -3.333333333333333D-002
  CONST B6  =  2.380952380952381D-002
  CONST B8  = -3.333333333333333D-002
  CONST B10 =  7.575757575757576D-002
  CONST B12 = -0.2531135531135531#

  IF X <= 0 THEN
    ErrCode = FSing
    TriGamma = MaxNum
    EXIT FUNCTION
  END IF

  DIM AS DOUBLE Res, Y, Z

  ErrCode = FOk
  Res = 0
  Z = X

  IF Z <= A THEN  ' Use small value approximation
    TriGamma = One / (Z * Z)
    EXIT FUNCTION
  END IF

  DO WHILE Z < B   ' Increase argument to (x+i) >= b
    Res = Res + One / (Z * Z)
    Z = Z + One
  LOOP

  ' Apply asymptotic formula where argument >= b
  Y = One / (Z * Z)
  Res = Res + Half * Y + (One + Y * (B2 + Y * (B4 + Y * (B6 + Y * _
                         (B8 + Y * (B10 + Y * B12)))))) / Z
  TriGamma = Res
END FUNCTION

