' ******************************************************************
'                 Gamma function and related functions
' Translated from C code in Cephes library (http://www.moshier.net)
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE FOk 0  
' No error

#DEFINE FSing -2  
' Singularity

#DEFINE FOverflow -3  
' Overflow range error

' ------------------------------------------------------------------
' Machine-dependent constants
' ------------------------------------------------------------------

#DEFINE MaxNum 1.797693134862315D+308   
' Max. floating point number: 2^1024

#DEFINE MaxGam 171.624376956302#        
' Max. argument for Gamma

' ------------------------------------------------------------------
' Mathematical constants
' ------------------------------------------------------------------

#DEFINE Pi 3.141592653589793#   
' Pi

#DEFINE Sqrt2Pi 2.506628274631001#   
' Sqrt(2*Pi)

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the latest function evaluation

' ******************************************************************

FUNCTION SgnGamma (X AS DOUBLE) AS INTEGER
IF X > 0 THEN
    SgnGamma = 1
ELSEIF INT(ABS(X)) MOD 2 = 1 THEN
    SgnGamma = 1
ELSE
    SgnGamma = -1
END IF
END FUNCTION

FUNCTION Stirling (X AS DOUBLE) AS DOUBLE
' Approximates Gamma function with Stirling's formula, for X >= 30

DIM AS DOUBLE P, W

W = 1 / X
P = 7.87311395793094D-04 * W - 2.29549961613378D-04
P = P * W - 2.68132617805781D-03
P = P * W + 3.47222221605459D-03
P = P * W + 8.33333333333482D-02
Stirling = Sqrt2Pi * EXP((X - 0.5) * LOG(X) - X) * (1 + W * P)
END FUNCTION

FUNCTION Gamma(X AS DOUBLE) AS DOUBLE

DIM AS INTEGER N, SgnGam
DIM AS DOUBLE  A, P, Q, X1, Z

SgnGam = SgnGamma(X)

IF X = 0 OR (X < 0 AND X = INT(X)) THEN
    ErrCode = FSing
    Gamma = SgnGam * MaxNum
    EXIT FUNCTION
END IF

IF X > MaxGam THEN
    ErrCode = FOverflow
    Gamma = MaxNum
    EXIT FUNCTION
END IF

ErrCode = FOk

A = ABS(X)
IF A >= 30 THEN
    IF X < 0 THEN
        N = INT(A)
        IF N = A THEN GOTO Goverf
        Z = A - N
        IF Z > .5 THEN N = N + 1: Z = A - N
        Z = ABS(A * SIN(Pi * Z))
        IF Z = 0 THEN GOTO Goverf
        Z = Pi / (Z * Stirling(A))
    ELSE
        Z = Stirling(X)
    END IF
    Gamma = SgnGam * Z
ELSE
    Z = 1
    X1 = X
    DO WHILE X1 >= 3
        X1 = X1 - 1
        Z = Z * X1
    LOOP
    DO WHILE X1 < 0
        IF X1 > -0.000000001# THEN GOTO Small
        Z = Z / X1
        X1 = X1 + 1
    LOOP
    DO WHILE X1 < 2
        IF X1 < 0.000000001# THEN GOTO Small
        Z = Z / X1
        X1 = X1 + 1
    LOOP
    IF X1 = 2 OR X1 = 3 THEN
        Gamma = Z
    ELSE
        X1 = X1 - 2
        P = 1.60119522476752D-04 * X1 + 1.19135147006586D-03
        P = P * X1 + 1.04213797561762D-02
        P = P * X1 + 4.76367800457137D-02
        P = P * X1 + .207448227648436#
        P = P * X1 + .494214826801497#
        P = P * X1 + 1
        Q = -2.3158187332412D-05 * X1 + 5.39605580493303D-04
        Q = Q * X1 - 4.45641913851797D-03
        Q = Q * X1 + .011813978522206#
        Q = Q * X1 + 3.58236398605499D-02
        Q = Q * X1 - .234591795718243#
        Q = Q * X1 + 7.14304917030273D-02
        Q = Q * X1 + 1
        Gamma = Z * P / Q
    END IF
END IF
EXIT FUNCTION

Goverf:
ErrCode = FOverflow
Gamma = SgnGam * MaxNum
EXIT FUNCTION

Small:
IF X1 = 0 THEN
    ErrCode = FSing
    Gamma = SgnGam * MaxNum
ELSE
    Gamma = Z / ((1 + 0.577215664901533# * X1) * X1)
END IF
END FUNCTION



