' ******************************************************************
' Gauss-Legendre integration
' ******************************************************************

CONST Npts = 8  ' Number of points / 2

DIM SHARED AS DOUBLE Root(1 TO Npts) = _
  {0.0950125098376370440185#, _
   0.281603550778258913230#,  _
   0.458016777657227386342#,  _
   0.617876244402643748447#,  _
   0.755404408355003033895#,  _
   0.865631202387831743880#,  _
   0.944575023073232576078#,  _
   0.989400934991649932596#}

DIM SHARED AS DOUBLE Weight(1 TO Npts) = _
  {0.189450610455068496285#, _
   0.182603415044923588867#, _
   0.169156519395002538189#, _
   0.149595988816576732081#, _
   0.124628971255533872052#, _
   0.095158511682492784810#, _
   0.062253523938647892863#, _
   0.027152459411754094852#}

' ******************************************************************

FUNCTION GausLeg(Func AS FUNCTION(X AS DOUBLE) AS DOUBLE,_
                 A    AS DOUBLE,_
                 B    AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Computes the integral of function Func from A to B
' by the Gauss-Legendre method
' ------------------------------------------------------------------

  DIM AS DOUBLE  P, Q, Sum, X, Y
  DIM AS INTEGER I
  
  P = 0.5 * (B + A)
  Q = 0.5 * (B - A)

  Sum = 0
  FOR I = 1 TO Npts
    X = Q * Root(I)
    Y = Func(P + X) + Func(P - X)
    Sum = Sum + Weight(I) * Y
  NEXT I

  GausLeg = Q * Sum
END FUNCTION

FUNCTION GausLeg0(Func AS FUNCTION(X AS DOUBLE) AS DOUBLE,_
                  B    AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Computes the integral of function Func from 0 to B 
' by the Gauss-Legendre method
' ------------------------------------------------------------------

  DIM AS DOUBLE  P, Sum, X, Y
  DIM AS INTEGER I

  P = 0.5 * B

  Sum = 0
  FOR I = 1 TO Npts
    X = P * Root(I)
    Y = Func(P + X) + Func(P - X)
    Sum = Sum + Weight(I) * Y
  NEXT I

  GausLeg0 = P * Sum
END FUNCTION

FUNCTION Convol(Func1 AS FUNCTION(X AS DOUBLE) AS DOUBLE,_
                Func2 AS FUNCTION(X AS DOUBLE) AS DOUBLE,_
                T     AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Computes the convolution product of two functions Func1 and Func2
' at time T by the Gauss-Legendre method
' ------------------------------------------------------------------

  DIM AS DOUBLE  P, PpX, PmX, Sum, X, Y
  DIM AS INTEGER I

  P = 0.5 * T

  Sum = 0
  FOR I = 1 TO Npts
    X = P * Root(I)
    PpX = P + X
    PmX = P - X
    Y = Func1(T - PpX) * Func2(PpX) + Func1(T - PmX) * Func2(PmX)
    Sum = Sum + Weight(I) * Y
  NEXT I

  Convol = P * Sum
END FUNCTION
