' ******************************************************************
' Quartic equation
' ******************************************************************

' ------------------------------------------------------------------
' Machine-dependent constant
' ------------------------------------------------------------------

#DEFINE MaxNum 1.797693134862315D+308   
' Max. floating point number: 2^1024

' ------------------------------------------------------------------
' Type definition
' ------------------------------------------------------------------

TYPE Complex
  X AS DOUBLE
  Y AS DOUBLE
END TYPE

' ------------------------------------------------------------------
' External function
' ------------------------------------------------------------------

DECLARE FUNCTION RootPol2(Coef() AS DOUBLE,_ 
                          Z()    AS Complex) AS INTEGER  
' Quadratic equation

DECLARE FUNCTION RootPol3(Coef() AS DOUBLE,_ 
                          Z()    AS Complex) AS INTEGER  
' Cubic equation

' ******************************************************************

FUNCTION HighestRealRoot(Deg AS INTEGER, Z() AS Complex) AS DOUBLE
' Returns highest real root 

  DIM R AS DOUBLE, I AS INTEGER
  
  R = -MaxNum
 
  FOR I = 1 TO Deg
    IF Z(I).Y = 0 AND Z(I).X > R THEN R = Z(I).X
  NEXT I

  HighestRealRoot = R  
END FUNCTION

FUNCTION RootPol4(Coef() AS DOUBLE, Z() AS Complex) AS INTEGER 
' ------------------------------------------------------------------
' Solves the quartic equation:
' Coef(0) + Coef(1) * Z + Coef(2) * Z^2 + Coef(3) * Z^3 +
'                                            Coef(4) * Z^4 = 0
' ------------------------------------------------------------------

  DIM AS INTEGER I, N, N1, N2
  DIM AS DOUBLE  A, AA, B, C, D, K, KK, L, M, Q, R, S, Cf(0 TO 3)
  DIM AS Complex Z1(1 TO 2), Z2(1 TO 2)

  FOR I = 1 TO 4
    Z(I).X = 0
    Z(I).Y = 0
  NEXT I

  IF Coef(4) = 0 THEN
    RootPol4 = RootPol3(Coef(), Z())
    EXIT FUNCTION
  END IF

  IF Coef(0) = 0 THEN
    ' 0 is root. Equation becomes cubic
    Cf(0) = Coef(1): Cf(1) = Coef(2): Cf(2) = Coef(3)

    ' Solve cubic equation
    RootPol4 = RootPol3(Cf(), Z()) + 1

    EXIT FUNCTION
  END IF

  IF Coef(4) = 1 THEN
    A = Coef(3) * 0.25
    B = Coef(2)
    C = Coef(1)
    D = Coef(0)
  ELSE
    A = Coef(3) / Coef(4) * 0.25
    B = Coef(2) / Coef(4)
    C = Coef(1) / Coef(4)
    D = Coef(0) / Coef(4)
  END IF

  AA = A * A

  Q = B - 6 * AA
  R = C + A * (8 * AA - 2 * B)
  S = D - A * C + AA * (B - 3 * AA)

  ' Compute coefficients of cubic equation
  Cf(3) = 1
  Cf(2) = 0.5 * Q
  Cf(1) = 0.25 * (Cf(2) * Cf(2) - S)

  ' Solve cubic equation and set KK = highest real root
  IF R = 0 AND Cf(1) < 0 THEN
    ' Eq. becomes quadratic with 2 real roots
    Cf(0) = Cf(1): Cf(1) = Cf(2): Cf(2) = 1
    N1 = RootPol2(Cf(), Z())
    KK = HighestRealRoot(2, Z())
  ELSE
    Cf(0) = -0.015625 * R * R
    N1 = RootPol3(Cf(), Z())
    KK = HighestRealRoot(3, Z())
  END IF

  K = SQR(KK)
  IF K = 0 THEN
    R = SQR(Q * Q - 4 * S)
  ELSE
    Q = Q + 4 * KK
    R = 0.5 * R / K
  END IF

  L = 0.5 * (Q - R)
  M = 0.5 * (Q + R)

  ' Solve quadratic equation: Z^2 + 2KZ + L = 0
  Cf(0) = L: Cf(1) = 2 * K: Cf(2) = 1
  N1 = RootPol2(Cf(), Z1())

  ' Solve quadratic equation: Z^2 - 2KZ + M = 0
  Cf(0) = M: Cf(1) = -Cf(1)
  N2 = RootPol2(Cf(), Z2())

  ' Transfer roots into vectors Xr and Xi
  Z(1).X = Z1(1).X - A : Z(1).Y = Z1(1).Y
  Z(2).X = Z1(2).X - A : Z(2).Y = Z1(2).Y
  Z(3).X = Z2(1).X - A : Z(3).Y = Z2(1).Y
  Z(4).X = Z2(2).X - A : Z(4).Y = Z2(2).Y

  RootPol4 = N1 + N2
END FUNCTION

