' ******************************************************************
'               Inverse of Normal distribution function
' Translated from C code in Cephes library (http://www.moshier.net)
' ******************************************************************

' ------------------------------------------------------------------
' Error codes
' ------------------------------------------------------------------

#DEFINE FOk 0  
' No error

#DEFINE FDomain -1  
' Argument domain error

' ------------------------------------------------------------------
' Machine-dependent constant
' ------------------------------------------------------------------

#DEFINE MaxNum 1.797693134862315D+308
' Max. floating point number: 2^1024

' ------------------------------------------------------------------
' Mathematical constant
' ------------------------------------------------------------------

#DEFINE Sqrt2Pi 2.506628274631001#   
' Sqrt(2*Pi)

' ------------------------------------------------------------------
' Global variable
' ------------------------------------------------------------------

COMMON SHARED ErrCode AS INTEGER
' Error code from the latest function evaluation

' ******************************************************************

FUNCTION PolEvl(X AS DOUBLE, Coef() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Evaluates polynomial of degree N:
'
'                       2          N
'   y  =  C  + C x + C x  +...+ C x
'          0    1     2          N
'
' Coefficients are stored in reverse order:
'
' Coef(0) = C  , ..., Coef(N) = C
'            N                   0
'
' The function P1Evl() assumes that Coef(N) = 1.0 and is
' omitted from the array. Its calling arguments are
' otherwise the same as PolEvl().
' ------------------------------------------------------------------

  DIM AS DOUBLE  Ans
  DIM AS INTEGER I

  Ans = Coef(0)
  FOR I = 1 TO UBOUND(Coef)
    Ans = Ans * X + Coef(I)
  NEXT I
  PolEvl = Ans
END FUNCTION

FUNCTION P1Evl(X AS DOUBLE, Coef() AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Evaluate polynomial when coefficient of X is 1.0,
' otherwise same as PolEvl.
' ------------------------------------------------------------------

  DIM AS DOUBLE  Ans
  DIM AS INTEGER I

  Ans = X + Coef(0)
  FOR I = 1 TO UBOUND(Coef)
    Ans = Ans * X + Coef(I)
  NEXT I
  P1Evl = Ans
END FUNCTION

FUNCTION InvNorm(P AS DOUBLE) AS DOUBLE
' ------------------------------------------------------------------
' Inverse of Normal distribution function
'
' Returns the argument, X, for which the area under the Gaussian
' probability density function (integrated from minus infinity to X)
' is equal to P.
' ------------------------------------------------------------------

  IF P <= 0 OR P >= 1 THEN
    ErrCode = FDomain
    InvNorm = SGN(P) * MaxNum
    EXIT FUNCTION
  END IF

  DIM AS DOUBLE P0(0 TO 4) = {  _
   -5.99633501014107895267E1,   _
    9.80010754185999661536E1,   _
   -5.66762857469070293439E1,   _
    1.39312609387279679503E1,   _
   -1.23916583867381258016E0}   

  DIM AS DOUBLE Q0(0 TO 7) = {  _
    1.95448858338141759834E0,   _
    4.67627912898881538453E0,   _
    8.63602421390890590575E1,   _
   -2.25462687854119370527E2,   _
    2.00260212380060660359E2,   _
   -8.20372256168333339912E1,   _
    1.59056225126211695515E1,   _
   -1.18331621121330003142E0}

  DIM AS DOUBLE P1(0 TO 8) = {  _
    4.05544892305962419923E0,   _
    3.15251094599893866154E1,   _
    5.71628192246421288162E1,   _
    4.40805073893200834700E1,   _
    1.46849561928858024014E1,   _
    2.18663306850790267539E0,   _
   -1.40256079171354495875E-1,  _
   -3.50424626827848203418E-2,  _
   -8.57456785154685413611E-4}

  DIM AS DOUBLE Q1(0 TO 7) = {  _
    1.57799883256466749731E1,   _
    4.53907635128879210584E1,   _
    4.13172038254672030440E1,   _
    1.50425385692907503408E1,   _
    2.50464946208309415979E0,   _
   -1.42182922854787788574E-1,  _
   -3.80806407691578277194E-2,  _
   -9.33259480895457427372E-4}

  DIM AS DOUBLE P2(0 TO 8) = {  _
    3.23774891776946035970E0,   _
    6.91522889068984211695E0,   _
    3.93881025292474443415E0,   _
    1.33303460815807542389E0,   _
    2.01485389549179081538E-1,  _
    1.23716634817820021358E-2,  _
    3.01581553508235416007E-4,  _
    2.65806974686737550832E-6,  _
    6.23974539184983293730E-9}

  DIM AS DOUBLE Q2(0 TO 7) = {  _
    6.02427039364742014255E0,   _
    3.67983563856160859403E0,   _
    1.37702099489081330271E0,   _
    2.16236993594496635890E-1,  _
    1.34204006088543189037E-2,  _
    3.28014464682127739104E-4,  _
    2.89247864745380683936E-6,  _
    6.79019408009981274425E-9}

  DIM AS DOUBLE  X, Y, Z, Y2, X0, X1
  DIM AS INTEGER Code

  ErrCode = FOk
  Code =1
  Y = P

  IF Y > (1.0 - 0.13533528323661269189) THEN ' 0.135... = exp(-2)
    Y = 1 - Y
    Code = 0
  END IF

  IF Y > 0.13533528323661269189 THEN
    Y = Y - 0.5
    Y2 = Y * Y
    X = Y + Y * (Y2 * PolEvl(Y2, P0()) / P1Evl(Y2, Q0()))
    X = X * Sqrt2Pi
    InvNorm = X 
    EXIT FUNCTION
  END IF

  X  = SQR(- 2 * LOG(Y))
  X0 = X - LOG(X) / X
  Z  = 1 / X

  IF X < 8 THEN
    X1 = Z * PolEvl(Z, P1()) / P1Evl(Z, Q1())
  ELSE
    X1 = Z * PolEvl(Z, P2()) / P1Evl(Z, Q2())
  END IF

  X = X0 - X1
  IF Code <> 0 THEN X = -X

  InvNorm = X
END FUNCTION
